"""Core Desktop class.

The Desktop is the central object in AUIMD. It manages components as below :

----------------------
|     Menu Bar       |
----------------------
|                    |
|                    |
|      Content       |
|     (widgets)      |
|                    |
|                    |
|                    |
----------------------
| On Screen keyboard |
|                    |
----------------------
|     Bottom Bar     |
----------------------

Widgets are AUIMD applications, they are usually shown in all the content space.
The desktop keeps track of the widgets in a list. The first widget is the
displayed widget.
The displayed widget is put in a scroll area, this enables to resize the widget
with an arbitrary size and be able to view all of its content with scrolling.
This simulate some sort of virtual space.

The desktop manages widgets life cycle :
- showWidget : show a widget, possibly adding it to the list of widget if not
               already in
- closeWidget : close a widget and remove it from the list of widgets. The
                next widget in the list is then shown
It is also possible to hide a widget with setWidgetVisibility.
The home widget is a little exception : it cannot be closed and the desktop has 
some particular knowledge of it. Appart from that, this is a regular widget.

Two kinds of widgets :
- those available in the core (home, mainmenu, exception)
- all other widgets
They share exactly the same structure but core widgets are mandatory, all other
are optional.
All widgets must inherit from AuimdWidgets AND a derived class of QWidget.
AuimdWidgets provides a basic applicative framework, whereas the QWidget derived
class provides graphical features. See core/auimd_widget.py for more.

The desktop also provides some utility function such as :
- open/save file dialog
- wallpaper management
- various plumbing

Basically most components will have a reference to the desktop, because it 
provides links between objects.

"""

from PyQt4 import QtCore, QtGui

import sys
import dbus
import dbus.service
import dbus.mainloop.qt

from menu_bar import *
from bottom_bar import *
from keyboard import *
from core.widgets.home import *
from core.widgets.exception import *

from file_open_save import *


def createDesktop(device, fullscreen, no_cursor, embed):
  """Create a class for creating a desktop instance.

  This indirection is used to choose between QtGui.QWidget QtGui.QX11EmbedWidget
  base class depending on the value of the embed argument.

  """

  if embed:
    base_class=QtGui.QX11EmbedWidget
  else:
    base_class=QtGui.QWidget

  class Desktop(base_class):

    class DBusService(dbus.service.Object):
      """Handles requests made on the service exposed on the session bus."""
      def __init__(self, desktop):
        dbus.service.Object.__init__(self, dbus.SessionBus(), "/Desktop")
        self.desktop=desktop

      @dbus.service.method(dbus_interface='net.pierrox.auimd.Desktop', in_signature='sas', out_signature='')
      def StartWidget(self, class_name, args):
        self.desktop.startWidget(class_name, *args)

    def __init__(self, device, fullscreen, no_cursor, embed):
      base_class.__init__(self, None)

      if fullscreen:
        self.setGeometry(QtGui.qApp.desktop().availableGeometry())
        self.setFixedSize(self.size())
      elif embed:
        self.embedInto(long(embed))

      # try to hide the cursor
      if no_cursor: 
        QtGui.qApp.setOverrideCursor(QtGui.QCursor(QtCore.Qt.BlankCursor))

      # load the device layer
      module=__import__("devices/"+device+"/device")
      self.device=module.Device()

      # check the screen size for possible desktop resize
      # (uses when no window manager present)
      if fullscreen:
        self.connect(QtGui.qApp.desktop(), QtCore.SIGNAL("resized(int)"), self.desktopResize)

      # create desktop components
      self.menu_bar=MenuBar(self)
      self.content=QtGui.QScrollArea(self)
      self.content.setFrameStyle(QtGui.QFrame.NoFrame)
      self.bottom_bar=BottomBar(self)
      self.keyboard=Keyboard(self)
      self.keyboard.hide()

      # the list of widgets
      self.widgets=[]

      # add the home widget and show it
      self.home=Home(self)
      self.showWidget(self.home)

      # add all desktop components in a vertical layout
      layout=QtGui.QVBoxLayout(self)
      layout.addWidget(self.menu_bar)
      layout.addWidget(self.content)
      layout.addWidget(self.keyboard)
      layout.addWidget(self.bottom_bar)
      layout.setMargin(0)
      layout.setSpacing(0)

      # load the list of directories in which to look for widgets
      self.loadWidgetsDirectories()

      # establish the list of discovered widgets
      self.loadWidgets()

      # exception tracebacks are shown in a widget
      sys.excepthook=self.exceptionHandler

      # finally exposes DBus services
      self.dbus_iface=self.DBusService(self)

    def exceptionHandler(self, exc_type, exc_value, exc_traceback):
      """Display exceptions in a widget."""
      self.showWidget(Exception(self, exc_value, exc_traceback))

    def desktopResize(self, screen):
      """Resize the desktop according to the screen size.

      This function is only called when in fullscreen mode.

      """
      r=QtGui.qApp.desktop().availableGeometry()
      self.setGeometry(r)
      self.setFixedSize(r.size())
      
    def resizeEvent(self, e=None):
      """Resize event are generated when in windowed mode."""
      widget=self.content.widget()
      if(widget): 
        self.setWidgetVirtualSize(widget)

    def showHideKeyboard(self, state=None):
      """Show or hide the keyboard, depending on its current state.

      If state is None, invert the state. Otherwise use the state value.

      """
      if(state==None):
        if(self.keyboard.isVisible()):
          state=0
        else:
          state=1
    
      if(state):
        self.keyboard.show()
      else:
        self.keyboard.hide()
      
    def startWidget(self, class_name, *args):
      """Start then show a widget by its class name, with optional arguments.
      
      If a class is single instance, and an instance is already running, it
      is simply put on top.
      
      """

      #if class_name not in self.widget_class_names:
      #  return False

      # get a class according to its name
      wc=self.widget_class_names[class_name]

      # check whether this is a mono or multi instances class, and if mono check for
      # an already running instance
      if wc.monoInstance():
        l=[x for x in self.widgets if x.__class__.__name__==class_name]
        if len(l)>0:
          w=l[0]
          w.restart(*args)
        else:
          w=wc(self, *args)
      else:
        w=wc(self, *args)

      # finally show the widget
      self.showWidget(w)


    def showWidget(self, widget):
      """Show a widget, and add it to the widgets list if not already in.
      
      In all cases the widget is put in front of the list. It is also resized
      according to its virtual size if needed. Showing a widget also shows its
      buttons.
      
      """
      if(widget in self.widgets):
        self.widgets.remove(widget)
      self.widgets.insert(0, widget)

      # replace the actually visible widget (if any) by the new one
      if(self.content.widget()):
        self.content.takeWidget()
      self.content.setWidget(widget)
      widget.show()

      # a widget can have a virtual size, larger than the view
      self.setWidgetVirtualSize(widget)

      widget.activateWindow()
      
      # enable the keyboard only if the widget accepts it
      self.bottom_bar.setButtons(widget.leftButton(), widget.rightButton(), widget.keyboardAllowed())
      if(not widget.keyboardAllowed()): self.showHideKeyboard(0)

      # update the menubar
      self.menu_bar.setCloseButtonEnabled(not self.content.widget() is self.home)
      self.menu_bar.selectWidget(widget)

    def closeCurrentWidget(self):
      """Close the current widget (the displayed widget) and show the next."""
      if len(self.widgets)==1: return
      self.closeWidget(self.content.widget())

    def closeWidget(self, old):
      """Close a given widget, and show the next."""
      if old not in self.widgets: return

      if not old.readyToClose(): return

      self.widgets.remove(old)
      self.menu_bar.closeWidget(old)

      self.showWidget(self.widgets[0])
      
      old.close()
      old.deleteLater()

    def setWidgetVisibility(self, widget, visibility):
      """Show or hide a widget without closing it."""
      self.menu_bar.setWidgetVisibility(widget, visibility)

    def updateWidgetIconTitle(self, widget):
      """Propagate widget's icon and title changes."""
      self.menu_bar.updateWidgetIconTitle(widget)

    def openSaveFile(self, type, opener):
      """Internal function used by openFile and saveFile."""
      w=FileOpenSave(self, type)
      w.setWidgetTitle(opener.widgetTitle()+" - "+w.widgetTitle())
      self.showWidget(w)
      self.setWidgetVisibility(opener, 0)
      f=w.openSaveFile()
      self.setWidgetVisibility(opener, 1)
      self.closeWidget(w)
      return f

    def openFile(self, opener):
      """Shown a file picker for opening.
      
      opener is the widget at the origin of the request. It will be hidden
      while the file is being picked.
      Return the filename or None.
      
      """
      return self.openSaveFile(0, opener)

    def saveFile(self, opener):
      """Shown a file picker for saving.
      
      opener is the widget at the origin of the request. It will be hidden
      while the file is being picked.
      Return the filename or None.
      
      """
      return self.openSaveFile(1, opener)

    def chooseDirectory(self, opener):
      """Shown a file picker for directory choosing.
      
      opener is the widget at the origin of the request. It will be hidden
      while the directory is being choosen.
      Return the directory name or None.
      
      """
      return self.openSaveFile(2, opener)

    def setWallpaper(self, path, scaled):
      """Set the home wallpaper attributes."""
      self.home.setWallpaper(path, scaled)

    def getWallpaper(self):
      """Retrieve the wallpaper of the home widget."""
      return self.home.wallpaper

    def setCurrentWidgetVirtualSize(self, virtual_size):
      """Change the virtual area of a widget.
      
      Using None as virtual_size means to use the maximum size but no more 
      (no scrollbars).
      
      """
      w=self.content.widget()
      w.virtual_size=virtual_size
      self.setWidgetVirtualSize(w)

    def getCurrentWidgetVirtualSize(self):
      return self.content.widget().virtual_size

    def setWidgetVirtualSize(self, widget):
      """Manage the virtual size of a widget (both containing scroll area and widget size itself)."""
      if widget.virtual_size:
        self.content.setWidgetResizable(0)
        widget.resize(widget.virtual_size)
      else:
        self.content.setWidgetResizable(1)
        widget.resize(self.content.geometry().size())

    def closeEvent(self, e):
      """Try to shut down things cleanly."""
      self.keyboard.close()
      for i in self.widgets: 
        i.close()

    def exit(self):
      """Request the application to terminate."""
      self.close()

    def loadWidgets(self, do_reload=False):
      """Look recursively in widgets directories folder for valid widgets classes.
      
      Search for packages, where __init__.py contains a AuimdWidgets tuple,
      giving the list of widgets provided by the package.
      
      """

      # hold classes of widgets that have been discovered (dict category -> list of classes)
      self.widget_classes={}

      # index of all widgets by name (dict name -> class)
      self.widget_class_names={}

      categories=("Graphics", "Games", "Multimedia", "Tools", "Network", "PIM", "Settings", "System", "Other")
      for c in categories:
        self.widget_classes[c]=[]

      for directory in self.widgets_directories:
        for dirpath, dirs, files in os.walk(directory):
          if "__init__.py" in files:
            if dirpath[0]=="/":
              sys.path.insert(0, os.path.dirname(dirpath))
              m=__import__(os.path.basename(dirpath))
            else:
              m=__import__(dirpath)
            try:
              if do_reload:
                reload(m)
                # this way to manually reload widgets is not completely satisfying because we need to list exhaustively all required modules for a widget
                modules=getattr(m, "AuimdModules", None)
                if modules:
                  for i in modules:
                    reload(i)
                # a second reload is needed because AuimdWidgets references modules before their reload
                reload(m)
              auimd_widgets=getattr(m, "AuimdWidgets")
              for w in auimd_widgets:
                c=w.getCategory()
                if not c in categories:
                  c="Other"
                self.widget_classes[c].append(w)
                self.widget_class_names[w.__name__]=w
                self.widget_packages.append(m)
            except AttributeError:
              # does not look like an auimd package
              pass

            # remove a search path that could have been inserted in cas of absolute path
            if dirpath[0]=="/":
              del sys.path[0]

    def reloadWidgets(self):
      """Reload widgets (after a first use of loadWidgets)."""
      self.loadWidgets(True)

    def loadWidgetsDirectories(self):
      self.widgets_directories=["widgets"]
      settings=QtCore.QSettings("auimd", "Desktop")
      n=settings.beginReadArray("widgets_directories")
      for i in range(n):
        settings.setArrayIndex(i)
        self.widgets_directories.append(str(settings.value("directory").toString()))
      settings.endArray()

    def saveWidgetsDirectories(self):
      settings=QtCore.QSettings("auimd", "Desktop")
      settings.beginWriteArray("widgets_directories")
      for i in range(len(self.widgets_directories)-1):
        settings.setArrayIndex(i)
        settings.setValue("directory", QtCore.QVariant(self.widgets_directories[i+1]))
      settings.endArray()

  return Desktop(device, fullscreen, no_cursor, embed)
